#This is a standalone version of testFSA for running on the web interface.
#It should only run on spinnaker.

#This is the class for construcing FSAs.  The individual CAs (both inputs
#and states) run at 5 ms.  There may be problems if they run faster or slower.
#The class enables you to build 2 input FSAs.  You can't go from state A
#to state B, then directly back.
#There are test functions at the end. 


import pyNN.spiNNaker as sim


#import numpy as np

DELAY = 1.0
SIM_LENGTH = 200.0

def nealProjection(preNeurons,postNeurons, connectorList,inhExc):
    connList = sim.FromListConnector(connectorList)
    sim.Projection(preNeurons, postNeurons, connList,target=inhExc)   

def init():
    print "spin" 
    sim.setup(timestep=DELAY,
                    min_delay=DELAY,
                    max_delay=DELAY, debug=0)


def createTwoInputs():
    inputSpikeTimes0 = [10.0]
    inputSpikeTimes1 = [50.0]
    spikeArray0 = {'spike_times': [inputSpikeTimes0]}
    spikeGen0=sim.Population(1,sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')
    spikeArray1 = {'spike_times': [inputSpikeTimes1]}
    spikeGen1=sim.Population(1, sim.SpikeSourceArray, spikeArray1,
                                   label='inputSpikes_1')

    return [spikeGen0,spikeGen1]

def createNeurons():
    numNeurons = 100
        
    cells=sim.Population(numNeurons,sim.IF_cond_exp,fsa.CELL_PARAMS)
    return cells


class FSAHelperFunctions:
    def __init__(self, simName):
        self.simulator = simName
        #self.neal = NealCoverFunctions(self.simulator)
        self.initParams()

    #FSA Parmaeters
    def initParams(self):
        self.CA_SIZE = 5  #This will (almost certainly) not work with a 
                     #different sized CA.

        self.INPUT_WEIGHT = 0.1
        self.INTRA_CA_WEIGHT = 0.03 
        self.HALF_ON_WEIGHT = 0.0017 
        self.FULL_ON_WEIGHT = 0.01
        self.CA_STOPS_CA_WEIGHT = -0.3 
        self.CELL_PARAMS = {'v_thresh':-55.0, 'v_reset' : -70.0, 
                                'tau_refrac': 2.0 , 'tau_syn_E': 5.0,  
                                #'e_rev_I': -100.0, 
                                #'tau_syn_I': 1.0, #bug2 fix
                                'v_rest' : -65.0,'i_offset':0.0}


    #--------Finite State Automata Functions ------------
    #---Functions that turn on states by one item

    #-- Function to ignite a state from a spike source
    #-- Uses INPUT_WEIGHT
    #use this when your using a spikesource
    def turnOnStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        connector = []
        for toOffset in range (0,self.CA_SIZE):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(0,toNeuron,self.INPUT_WEIGHT,
                                      DELAY)]
        nealProjection(spikeSource, toNeurons, connector,'excitatory')

    def turnOnStateFromOneNeuron(self,fromNeurons,fromNeuron, toNeurons, toCA):
        connector = []
        for toOffset in range (0,self.CA_SIZE):
            toNeuron = toOffset + (toCA*self.CA_SIZE)
            connector = connector + [(fromNeuron,toNeuron,self.INPUT_WEIGHT,
                                      DELAY)]
        nealProjection(fromNeurons, toNeurons, connector,'excitatory')

    #Function to turn on one state from another
    #Call with fromPopulation, fromCA, toPopulation and toCA
    def turnOnStateFromState(self,fromNeurons,fromCA,toNeurons,toCA):
        connector = []
        #uses FULL_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.FULL_ON_WEIGHT,
                                          DELAY)]

        nealProjection(fromNeurons,toNeurons,connector,'excitatory')


    #---Create a CA that will persistently fire.
    #-- Assumes neurons in the same population
    #-- Uses INTRA_CA_WEIGHT
    def makeCA(self,neurons, CA):
        #print 'makeCA
        connector = []
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (CA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (CA*self.CA_SIZE)
                if (toNeuron != fromNeuron):
                    connector = connector + [(fromNeuron,toNeuron,
                                              self.INTRA_CA_WEIGHT,
                                              DELAY)]

        nealProjection(neurons,neurons,connector,'excitatory')

    #--Functions when two inputs are needed to turn on a third
    #-- Two states are needed to turn on a third.  
    #-- This connects one of the inputs to the the third.
    #-- Uses HALF_ON_WEIGHT
    def stateHalfTurnsOnState(self,fromNeurons,fromCA,toNeurons,toCA):
        connector = []
        #uses HALF_ON_WEIGHT
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector=connector+[(fromNeuron,toNeuron,
                                          self.HALF_ON_WEIGHT,
                                          DELAY)]

        nealProjection(fromNeurons,toNeurons,connector,'excitatory')

    #---- One State or other set of neurons turns off another
    #-- Uses CA_STOPS_CA_WEIGHT
    def stateTurnsOffState(self,fromNeurons, fromCA, toNeurons, toCA):
        connector = []
        for fromOffset in range (0,self.CA_SIZE):
            fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
            for toOffset in range (0,self.CA_SIZE):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,
                                              self.CA_STOPS_CA_WEIGHT,
                                              DELAY)]

        nealProjection(fromNeurons,toNeurons,connector,'inhibitory')


#------test functions 
    #in class version
    def test3StateFSA(self, firstSpikeGenerator, secondSpikeGenerator,
                      stateCells):
        #Build the FSA
        self.turnOnStateFromSpikeSource(firstSpikeGenerator,stateCells,0)
        self.turnOnStateFromSpikeSource(secondSpikeGenerator,stateCells,1)
        self.makeCA(stateCells,0)
        self.makeCA(stateCells,1)
        self.makeCA(stateCells,2)
        self.stateHalfTurnsOnState(stateCells,0,stateCells,2)
        self.stateTurnsOffState(stateCells,2,stateCells,0)
        #comment below out to check state 0 alone does not turn on state 2
        self.stateHalfTurnsOnState(stateCells,1,stateCells,2)
        self.stateTurnsOffState(stateCells,2,stateCells,1)

#---main

fsa = FSAHelperFunctions('spinnaker')
init()
spikeGenerators = createTwoInputs()
firstSpikeGenerator =  spikeGenerators[0]
secondSpikeGenerator =  spikeGenerators[1]

stateCells = createNeurons()

stateCells.record()

#Build the FSA
fsa.turnOnStateFromSpikeSource(firstSpikeGenerator,stateCells,0)
fsa.turnOnStateFromSpikeSource(secondSpikeGenerator,stateCells,1)
fsa.makeCA(stateCells,0)
fsa.makeCA(stateCells,1)
fsa.makeCA(stateCells,2)
fsa.stateHalfTurnsOnState(stateCells,0,stateCells,2)
fsa.stateTurnsOffState(stateCells,2,stateCells,0)
#comment below out to check state 0 alone does not turn on state 2
fsa.stateHalfTurnsOnState(stateCells,1,stateCells,2)
fsa.stateTurnsOffState(stateCells,2,stateCells,1)

sim.run(SIM_LENGTH)

stateCells.printSpikes('temp.sp')
