#This makes a 2 state FSA.  It does it without calling any of the 
#FSA test functions.  It should have the same results as testFSASmall.py.
#State 0, with input 1, turns on state 2.  2 turns off 1 and 0.

import nealParams as nealParameters
#get the simulator
if (nealParameters.pynnVersion=="pynn")and(nealParameters.simulator=="spinnaker"):
    import pyNN.spiNNaker as sim
elif (nealParameters.pynnVersion=="pynn")and(nealParameters.simulator=="nest"):
    import pyNN.nest as sim
    import cPickle as pickle
elif (nealParameters.pynnVersion=="pynest")and(nealParameters.simulator=="nest"):
    import nest as nest

import numpy as np

from nealCoverClass import NealCoverFunctions
from FSAClass import FSAHelperFunctions

#---Below here are functions to test the system---
RUN_DURATION = 200
#initialize the simulator. 
def init():
    if (nealParameters.pynnVersion == "pynn"):
        #print "spin" or nest
        sim.setup(timestep=nealParameters.DELAY,
                    min_delay=nealParameters.DELAY,
                    max_delay=nealParameters.DELAY, debug=0)

    elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        nest.ResetKernel()
        nest.SetKernelStatus({'resolution':nealParameters.DELAY})


def printPklSpikes(fileName):
    fileHandle = open(fileName)
    neoObj = pickle.load(fileHandle)
    segments = neoObj.segments
    segment = segments[0]
    spikeTrains = segment.spiketrains
    neurons = len(spikeTrains)
    for neuronNum in range (0,neurons):
        if (len(spikeTrains[neuronNum])>0):
            spikes = spikeTrains[neuronNum]
            for spike in range (0,len(spikes)):
                print neuronNum, spikes[spike]
    fileHandle.close()
    

def createTwoInputs():
    inputSpikeTimes0 = [10.0]
    inputSpikeTimes1 = [50.0]
    if (nealParameters.pynnVersion == "pynn"):
        spikeArray0 = {'spike_times': [inputSpikeTimes0]}
        spikeGen0=sim.Population(1,sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')
        spikeArray1 = {'spike_times': [inputSpikeTimes1]}
        spikeGen1=sim.Population(1, sim.SpikeSourceArray, spikeArray1,
                                   label='inputSpikes_1')

    elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        spikeGen0 = nest.Create('spike_generator',
            #params = {'spike_times': np.array([10.0,15.0, 20.0,25.0,30.0])})
            params = {'spike_times': np.array(inputSpikeTimes0)})

        spikeGen1 = nest.Create('spike_generator',
            params = {'spike_times': np.array(inputSpikeTimes1)})

        spikeDet = nest.Create('spike_detector')

    else: print "bad simulator for spike generator"

    return [spikeGen0,spikeGen1]


def createNeurons():
    if (nealParameters.pynnVersion == "pynn"): 
        if (nealParameters.simulator == 'nest'):
            numNeurons = 15
        elif (nealParameters.simulator == 'spinnaker'):
            numNeurons = 100
        
        cells=sim.Population(numNeurons,sim.IF_cond_exp,fsa.CELL_PARAMS)
    elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        #in python Models() to see all the models including these
        #print GetDefaults (neuronType)

        cells = nest.Create("iaf_cond_exp",n=15,params = fsa.CELL_PARAMS)


    return cells

def createRecorder():
    spikeDet = 0
    if (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        spikeDet = nest.Create('spike_detector')

    #with spinnaker you can just set record on the cells
    return spikeDet

def setupRecording(cells, spikeDetector):
    if (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'nest'):
        cells.record({'spikes','v'})
    elif (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'spinnaker'):
        cells.record()
    elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        nest.Connect(cells,spikeDetector)

def runFSA():
    if (nealParameters.pynnVersion == "pynn"):
        sim.run(RUN_DURATION)
    elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        nest.Simulate(RUN_DURATION)


def printResults(simCells,nestRecorder):
    #print
    if (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'nest'):
        simCells.printSpikes('temp.pkl')
        printPklSpikes('temp.pkl')
    elif (nealParameters.pynnVersion == "pynn") and (nealParameters.simulator == 'spinnaker'):
        simCells.printSpikes('temp.sp')
    elif (nealParameters.pynnVersion=="pynest") and (nealParameters.simulator == 'nest'):
        spikes = nest.GetStatus(nestRecorder)
        test = spikes[0]['events']['times']
        print 'spikes', test

##--main---
neal = NealCoverFunctions(nealParameters.simulator)
fsa = FSAHelperFunctions(nealParameters.simulator)
init()
spikeGenerators = createTwoInputs()
firstSpikeGenerator =  spikeGenerators[0]
secondSpikeGenerator =  spikeGenerators[1]

stateCells = createNeurons()

recorder = createRecorder()
setupRecording(stateCells,recorder)

#Build the FSA
fsa.turnOnStateFromSpikeSource(firstSpikeGenerator,stateCells,0)
fsa.turnOnStateFromSpikeSource(secondSpikeGenerator,stateCells,1)
fsa.makeCA(stateCells,0)
fsa.makeCA(stateCells,1)
fsa.makeCA(stateCells,2)
fsa.stateHalfTurnsOnState(stateCells,0,stateCells,2)
fsa.stateTurnsOffState(stateCells,2,stateCells,0)
#comment below out to check state 0 alone does not turn on state 2
fsa.stateHalfTurnsOnState(stateCells,1,stateCells,2)
fsa.stateTurnsOffState(stateCells,2,stateCells,1)

runFSA()

printResults(stateCells,recorder)



